   /// Says hello.
//    ┡━━━━━━━━━━┛╰─ markdown
//    ╰─ markdown markdown-inline
   ///
//    ╰─ markdown
   /// # Example
//    ┡━┛┡━━━━━┛╰─ markdown
//    │  ╰─ markdown markdown-inline
//    ╰─ markdown
   ///
//    ╰─ markdown
   /// ```rust
//    ┗━━━━━━┹─ markdown
   /// fn add(left: usize, right: usize) -> usize {
//    ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┹─ markdown rust
   ///     left + right
//    ┗━━━━━━━━━━━━━━━┹─ markdown rust
   /// }
//    ┗┹─ markdown rust
   /// ```
//    ┗━━┹─ markdown
   pub fn hello() {}
