   /// Says hello.
// ┡┛╿┗━━━━━━━━━━┹─ comment
// │ ╰─ comment comment
// ╰─ comment
   ///
// ┡┛╿╰─ comment
// │ ╰─ comment comment
// ╰─ comment
   /// # Example
// ┡┛╿┡┛┗━━━━━━┹─ comment markup.heading.1
// │ │╰─ comment markup.heading.1 markup.heading.marker
// │ ╰─ comment comment
// ╰─ comment
   ///
// ┡┛╿╰─ comment
// │ ╰─ comment comment
// ╰─ comment
   /// ```rust
// ┡┛╿┡━━┛┡━━┛╰─ comment markup.raw.block
// │ ││   ╰─ comment markup.raw.block label
// │ │╰─ comment markup.raw.block punctuation.bracket
// │ ╰─ comment comment
// ╰─ comment
   /// fn add(left: usize, right: usize) -> usize {
// ┡┛╿╿┡┛╿┡━┛╿┡━━┛╿╿┡━━━┛╿╿┡━━━┛╿╿┡━━━┛╿╿┡┛╿┡━━━┛╿╿╰─ comment markup.raw.block
// │ │││ ││  ││   │││    │││    │││    │││ ││    │╰─ comment markup.raw.block punctuation.bracket
// │ │││ ││  ││   │││    │││    │││    │││ ││    ╰─ comment markup.raw.block
// │ │││ ││  ││   │││    │││    │││    │││ │╰─ comment markup.raw.block type.builtin
// │ │││ ││  ││   │││    │││    │││    │││ ╰─ comment markup.raw.block
// │ │││ ││  ││   │││    │││    │││    ││╰─ comment markup.raw.block operator
// │ │││ ││  ││   │││    │││    │││    │╰─ comment markup.raw.block
// │ │││ ││  ││   │││    │││    │││    ╰─ comment markup.raw.block punctuation.bracket
// │ │││ ││  ││   │││    │││    ││╰─ comment markup.raw.block type.builtin
// │ │││ ││  ││   │││    │││    │╰─ comment markup.raw.block
// │ │││ ││  ││   │││    │││    ╰─ comment markup.raw.block punctuation.delimiter
// │ │││ ││  ││   │││    ││╰─ comment markup.raw.block variable.parameter
// │ │││ ││  ││   │││    │╰─ comment markup.raw.block
// │ │││ ││  ││   │││    ╰─ comment markup.raw.block punctuation.delimiter
// │ │││ ││  ││   ││╰─ comment markup.raw.block type.builtin
// │ │││ ││  ││   │╰─ comment markup.raw.block
// │ │││ ││  ││   ╰─ comment markup.raw.block punctuation.delimiter
// │ │││ ││  │╰─ comment markup.raw.block variable.parameter
// │ │││ ││  ╰─ comment markup.raw.block punctuation.bracket
// │ │││ │╰─ comment markup.raw.block function
// │ │││ ╰─ comment markup.raw.block
// │ ││╰─ comment markup.raw.block keyword.function
// │ │╰─ comment markup.raw.block
// │ ╰─ comment comment
// ╰─ comment
   ///     left + right
// ┡┛╿┡━━━┛┡━━┛╿╿╿┡━━━┛╰─ comment markup.raw.block
// │ ││    │   │││╰─ comment markup.raw.block variable.parameter
// │ ││    │   ││╰─ comment markup.raw.block
// │ ││    │   │╰─ comment markup.raw.block operator
// │ ││    │   ╰─ comment markup.raw.block
// │ ││    ╰─ comment markup.raw.block variable.parameter
// │ │╰─ comment markup.raw.block
// │ ╰─ comment comment
// ╰─ comment
   /// }
// ┡┛╿╿╿╰─ comment markup.raw.block
// │ ││╰─ comment markup.raw.block punctuation.bracket
// │ │╰─ comment markup.raw.block
// │ ╰─ comment comment
// ╰─ comment
   /// ```
// ┡┛╿┡━━┛╰─ comment markup.raw.block
// │ │╰─ comment markup.raw.block punctuation.bracket
// │ ╰─ comment comment
// ╰─ comment
   pub fn hello() {}
// ┡━┛ ┡┛ ┡━━━┛┡┛ ┗┹─ punctuation.bracket
// │   │  │    ╰─ punctuation.bracket
// │   │  ╰─ function
// │   ╰─ keyword.function
// ╰─ keyword
