   // TODO **hello-world** 
// ┡━┛┡━━┛┗━━━━━━━━━━━━━━━┹─ comment
// │  ╰─ comment info
// ╰─ comment
   fn foo() {
// ┡┛ ┡━┛┡┛ ╰─ punctuation.bracket
// │  │  ╰─ punctuation.bracket
// │  ╰─ function
// ╰─ keyword.function
       println!("hello world")
//     ┡━━━━━━┛╿┡━━━━━━━━━━━┛╰─ punctuation.bracket
//     │       │╰─ string
//     │       ╰─ punctuation.bracket
//     ╰─ function.macro
   }
// ╰─ punctuation.bracket
   /// bar**
// ┗━━━━━━━┹─ comment
   fn bar() {
// ┡┛ ┡━┛┡┛ ╰─ punctuation.bracket
// │  │  ╰─ punctuation.bracket
// │  ╰─ function
// ╰─ keyword.function
       println!("hello world")
//     ┡━━━━━━┛╿┡━━━━━━━━━━━┛╰─ punctuation.bracket
//     │       │╰─ string
//     │       ╰─ punctuation.bracket
//     ╰─ function.macro
   }
// ╰─ punctuation.bracket
