   fn add(left: usize, right: usize) -> usize {
// ┡┛ ┡━┛╿┡━━┛╿ ┡━━━┛╿ ┡━━━┛╿ ┡━━━┛╿ ┡┛ ┡━━━┛ ╰─ punctuation.bracket
// │  │  ││   │ │    │ │    │ │    │ │  ╰─ type.builtin
// │  │  ││   │ │    │ │    │ │    │ ╰─ operator
// │  │  ││   │ │    │ │    │ │    ╰─ punctuation.bracket
// │  │  ││   │ │    │ │    │ ╰─ type.builtin
// │  │  ││   │ │    │ │    ╰─ punctuation.delimiter
// │  │  ││   │ │    │ ╰─ variable.parameter
// │  │  ││   │ │    ╰─ punctuation.delimiter
// │  │  ││   │ ╰─ type.builtin
// │  │  ││   ╰─ punctuation.delimiter
// │  │  │╰─ variable.parameter
// │  │  ╰─ punctuation.bracket
// │  ╰─ function
// ╰─ keyword.function
    left + right
//  ┡━━┛ ╿ ┗━━━┹─ variable.parameter
//  │    ╰─ operator
//  ╰─ variable.parameter
   }
// ╰─ punctuation.bracket
   
   fn add(left: usize, right: usize) -> usize {
// ┡┛ ┡━┛╿┡━━┛╿ ┡━━━┛╿ ┡━━━┛╿ ┡━━━┛╿ ┡┛ ┡━━━┛ ╰─ punctuation.bracket
// │  │  ││   │ │    │ │    │ │    │ │  ╰─ type.builtin
// │  │  ││   │ │    │ │    │ │    │ ╰─ operator
// │  │  ││   │ │    │ │    │ │    ╰─ punctuation.bracket
// │  │  ││   │ │    │ │    │ ╰─ type.builtin
// │  │  ││   │ │    │ │    ╰─ punctuation.delimiter
// │  │  ││   │ │    │ ╰─ variable.parameter
// │  │  ││   │ │    ╰─ punctuation.delimiter
// │  │  ││   │ ╰─ type.builtin
// │  │  ││   ╰─ punctuation.delimiter
// │  │  │╰─ variable.parameter
// │  │  ╰─ punctuation.bracket
// │  ╰─ function
// ╰─ keyword.function
    left + right
//  ┡━━┛ ╿ ┗━━━┹─ variable.parameter
//  │    ╰─ operator
//  ╰─ variable.parameter
   }
// ╰─ punctuation.bracket
