   fn main() {
// ┡┛ ┡━━┛┡┛ ╰─ punctuation.bracket
// │  │   ╰─ punctuation.bracket
// │  ╰─ function
// ╰─ keyword.function
       println!("hello world")
//     ┡━━━━━━┛╿┡━━━━━━━━━━━┛╰─ punctuation.bracket
//     │       │╰─ string
//     │       ╰─ punctuation.bracket
//     ╰─ function.macro
   }
// ╰─ punctuation.bracket
