   %% <li>foo
// ┡━┛╿┡┛╿┗━┹─ comment
// │  ││ ╰─ comment punctuation.bracket
// │  │╰─ comment tag
// │  ╰─ comment punctuation.bracket
// ╰─ comment
   %% bar</li>
// ┡━━━━┛┡┛┡┛╿╰─ comment
// │     │ │ ╰─ comment punctuation.bracket
// │     │ ╰─ comment tag
// │     ╰─ comment punctuation.bracket
// ╰─ comment
